/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _ScriptChildProcess_instances, _ScriptChildProcess_script, _ScriptChildProcess_child, _ScriptChildProcess_state, _ScriptChildProcess_actuallyKill, _ScriptChildProcess_pathEnvironmentVariable_get;
import * as pathlib from 'path';
import { spawn } from 'child_process';
import { augmentProcessEnvSafelyIfOnWindows, IS_WINDOWS, } from './util/windows.js';
/**
 * The PATH environment variable of this process, minus all of the leading
 * "node_modules/.bin" entries that the incoming "npm run" command already set.
 *
 * We want full control over which "node_modules/.bin" paths are in the PATH of
 * the processes we spawn, so that cross-package dependencies act as though we
 * are running "npm run" with each package as the cwd.
 *
 * We only need to do this once per Wireit process, because process.env never
 * changes.
 */
const PATH_ENV_SUFFIX = (() => {
    const path = process.env.PATH ?? '';
    // Note the PATH delimiter is platform-dependent.
    const entries = path.split(pathlib.delimiter);
    const nodeModulesBinSuffix = pathlib.join('node_modules', '.bin');
    const endOfNodeModuleBins = entries.findIndex((entry) => !entry.endsWith(nodeModulesBinSuffix));
    return entries.slice(endOfNodeModuleBins).join(pathlib.delimiter);
})();
/**
 * A child process spawned during execution of a script.
 */
export class ScriptChildProcess {
    constructor(script) {
        _ScriptChildProcess_instances.add(this);
        _ScriptChildProcess_script.set(this, void 0);
        _ScriptChildProcess_child.set(this, void 0);
        _ScriptChildProcess_state.set(this, 'starting');
        /**
         * Resolves when this child process ends.
         */
        Object.defineProperty(this, "completed", {
            enumerable: true,
            configurable: true,
            writable: true,
            value: void 0
        });
        __classPrivateFieldSet(this, _ScriptChildProcess_script, script, "f");
        // TODO(aomarks) Update npm_ environment variables to reflect the new
        // package.
        __classPrivateFieldSet(this, _ScriptChildProcess_child, spawn(script.command.value, script.extraArgs, {
            cwd: script.packageDir,
            // Conveniently, "shell:true" has the same shell-selection behavior as
            // "npm run", where on macOS and Linux it is "sh", and on Windows it is
            // %COMSPEC% || "cmd.exe".
            //
            // References:
            //   https://nodejs.org/api/child_process.html#child_processspawncommand-args-options
            //   https://nodejs.org/api/child_process.html#default-windows-shell
            //   https://github.com/npm/run-script/blob/a5b03bdfc3a499bf7587d7414d5ea712888bfe93/lib/make-spawn-args.js#L11
            shell: true,
            env: augmentProcessEnvSafelyIfOnWindows({
                PATH: __classPrivateFieldGet(this, _ScriptChildProcess_instances, "a", _ScriptChildProcess_pathEnvironmentVariable_get),
            }),
            // Set "detached" on Linux and macOS so that we create a new process
            // group, instead of being added to the process group for this Wireit
            // process.
            //
            // We need a new process group so that we can use "kill(-pid)" to kill all
            // of the processes in the process group, instead of just the group leader
            // "sh" process. "sh" does not forward signals to child processes, so a
            // regular "kill(pid)" would not kill the actual process we care about.
            //
            // On Windows this works differently, and we use the "\t" flag to
            // "taskkill" to kill child processes. However, if we do set "detached" on
            // Windows, it causes the child process to open in a new terminal window,
            // which we don't want.
            detached: !IS_WINDOWS,
        }), "f");
        this.completed = new Promise((resolve, reject) => {
            __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").on('spawn', () => {
                switch (__classPrivateFieldGet(this, _ScriptChildProcess_state, "f")) {
                    case 'starting': {
                        __classPrivateFieldSet(this, _ScriptChildProcess_state, 'started', "f");
                        break;
                    }
                    case 'killing': {
                        // We received a kill request while we were still starting. Kill now
                        // that we're started.
                        __classPrivateFieldGet(this, _ScriptChildProcess_instances, "m", _ScriptChildProcess_actuallyKill).call(this);
                        break;
                    }
                    case 'started':
                    case 'stopped': {
                        reject(new Error(`Internal error: Expected ScriptChildProcessState ` +
                            `to be "started" or "killing" but was "${__classPrivateFieldGet(this, _ScriptChildProcess_state, "f")}"`));
                        break;
                    }
                    default: {
                        const never = __classPrivateFieldGet(this, _ScriptChildProcess_state, "f");
                        reject(new Error(`Internal error: unexpected ScriptChildProcessState: ${String(never)}`));
                    }
                }
            });
            __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").on('error', (error) => {
                resolve({
                    ok: false,
                    error: {
                        script,
                        type: 'failure',
                        reason: 'spawn-error',
                        message: error.message,
                    },
                });
                __classPrivateFieldSet(this, _ScriptChildProcess_state, 'stopped', "f");
            });
            __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").on('close', (status, signal) => {
                if (__classPrivateFieldGet(this, _ScriptChildProcess_state, "f") === 'killing') {
                    resolve({
                        ok: false,
                        error: {
                            script,
                            type: 'failure',
                            reason: 'killed',
                        },
                    });
                }
                else if (signal !== null) {
                    resolve({
                        ok: false,
                        error: {
                            script,
                            type: 'failure',
                            reason: 'signal',
                            signal,
                        },
                    });
                }
                else if (status !== 0) {
                    resolve({
                        ok: false,
                        error: {
                            script,
                            type: 'failure',
                            reason: 'exit-non-zero',
                            // status should only ever be null if signal was not null, but
                            // this isn't reflected in the TypeScript types. Just in case, and
                            // to make TypeScript happy, fall back to -1 (which is a
                            // conventional exit status used for "exited with signal").
                            status: status ?? -1,
                        },
                    });
                }
                else {
                    resolve({ ok: true, value: undefined });
                }
                __classPrivateFieldSet(this, _ScriptChildProcess_state, 'stopped', "f");
            });
        });
    }
    get stdout() {
        return __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").stdout;
    }
    get stderr() {
        return __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").stderr;
    }
    /**
     * Kill this child process. On Linux/macOS, sends a `SIGINT` signal. On
     * Windows, invokes `taskkill /pid PID /t`.
     *
     * Note this function returns immediately. To find out when the process was
     * actually killed, use the {@link completed} promise.
     */
    kill() {
        switch (__classPrivateFieldGet(this, _ScriptChildProcess_state, "f")) {
            case 'started': {
                __classPrivateFieldGet(this, _ScriptChildProcess_instances, "m", _ScriptChildProcess_actuallyKill).call(this);
                return;
            }
            case 'starting': {
                // We're still starting up, and it's not possible to abort. When we get
                // the "spawn" event, we'll notice the "killing" state and actually kill
                // then.
                __classPrivateFieldSet(this, _ScriptChildProcess_state, 'killing', "f");
                return;
            }
            case 'killing':
            case 'stopped': {
                // No-op.
                return;
            }
            default: {
                const never = __classPrivateFieldGet(this, _ScriptChildProcess_state, "f");
                throw new Error(`Internal error: unexpected ScriptChildProcessState: ${String(never)}`);
            }
        }
    }
}
_ScriptChildProcess_script = new WeakMap(), _ScriptChildProcess_child = new WeakMap(), _ScriptChildProcess_state = new WeakMap(), _ScriptChildProcess_instances = new WeakSet(), _ScriptChildProcess_actuallyKill = function _ScriptChildProcess_actuallyKill() {
    if (__classPrivateFieldGet(this, _ScriptChildProcess_child, "f").pid === undefined) {
        throw new Error(`Internal error: Can't kill child process because it has no pid. ` +
            `Command: ${JSON.stringify(__classPrivateFieldGet(this, _ScriptChildProcess_script, "f").command)}.`);
    }
    if (IS_WINDOWS) {
        // Windows doesn't have signals. Node ChildProcess.kill() sort of emulates
        // the behavior of SIGKILL (and ignores the signal you pass in), but this
        // doesn't end child processes. We have child processes because the parent
        // process is the shell (cmd.exe or PowerShell).
        // https://docs.microsoft.com/en-us/windows-server/administration/windows-commands/taskkill
        spawn('taskkill', [
            '/pid',
            __classPrivateFieldGet(this, _ScriptChildProcess_child, "f").pid.toString(),
            /* End child processes */ '/t',
            /* Force. Killing does not seem reliable otherwise. */ '/f',
        ]);
    }
    else {
        // We used "detached" when we spawned, so our child is the leader of a
        // process group. Passing the negative of a pid kills all processes in
        // that group (without the negative, only the leader "sh" process would be
        // killed).
        process.kill(-__classPrivateFieldGet(this, _ScriptChildProcess_child, "f").pid, 'SIGINT');
    }
    __classPrivateFieldSet(this, _ScriptChildProcess_state, 'killing', "f");
}, _ScriptChildProcess_pathEnvironmentVariable_get = function _ScriptChildProcess_pathEnvironmentVariable_get() {
    // Given package "/foo/bar", walk up the path hierarchy to generate
    // "/foo/bar/node_modules/.bin:/foo/node_modules/.bin:/node_modules/.bin".
    const entries = [];
    let cur = __classPrivateFieldGet(this, _ScriptChildProcess_script, "f").packageDir;
    while (true) {
        entries.push(pathlib.join(cur, 'node_modules', '.bin'));
        const parent = pathlib.dirname(cur);
        if (parent === cur) {
            break;
        }
        cur = parent;
    }
    // Add the inherited PATH variable, minus any "node_modules/.bin" entries
    // that were set by the "npm run" command that spawned Wireit.
    entries.push(PATH_ENV_SUFFIX);
    // Note the PATH delimiter is platform-dependent.
    return entries.join(pathlib.delimiter);
};
//# sourceMappingURL=script-child-process.js.map