/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
export function computeManifestEntry(stats) {
    // We follow the same strategy as Git, see
    // https://github.com/git/git/blob/2668e3608e47494f2f10ef2b6e69f08a84816bcb/Documentation/technical/racy-git.txt#L34.
    return {
        t: stats.isFile()
            ? 'f'
            : stats.isDirectory()
                ? 'd'
                : stats.isSymbolicLink()
                    ? 'l'
                    : stats.isBlockDevice()
                        ? 'b'
                        : stats.isCharacterDevice()
                            ? 'c'
                            : stats.isFIFO()
                                ? 'F'
                                : stats.isSocket()
                                    ? 's'
                                    : '?',
        // Don't include timestamps for directories, because they change when a
        // child is added or removed. If we are tracking the child, then it will
        // have its own entry. If we are not tracking the child, then we don't want
        // it to affect the manifest.
        m: stats.isDirectory() ? null : stats.mtimeMs,
        c: stats.isDirectory() ? null : stats.ctimeMs,
        s: stats.size,
        u: stats.uid,
        g: stats.gid,
        i: stats.ino,
        M: stats.mode,
    };
}
//# sourceMappingURL=manifest.js.map