/**
 * @license
 * Copyright 2022 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Watcher_instances, _Watcher_state, _Watcher_rootScript, _Watcher_extraArgs, _Watcher_logger, _Watcher_workerPool, _Watcher_cache, _Watcher_failureMode, _Watcher_abort, _Watcher_latestRootScriptConfig, _Watcher_configFilesWatcher, _Watcher_inputFileWatchers, _Watcher_finished, _Watcher_startRun, _Watcher_analyze, _Watcher_execute, _Watcher_onRunDone, _Watcher_onConfigFileChanged, _Watcher_fileChanged, _Watcher_synchronizeInputFileWatchers, _Watcher_onAbort, _Watcher_closeAllFileWatchers;
import chokidar from 'chokidar';
import { Analyzer } from './analyzer.js';
import { Executor } from './executor.js';
import { Deferred } from './util/deferred.js';
import { scriptReferenceToString, } from './script.js';
function unknownState(state) {
    throw new Error(`Unknown watcher state ${String(state)}`);
}
function unexpectedState(state) {
    throw new Error(`Unexpected watcher state ${state}`);
}
/**
 * Watches a script for changes in its input files, and in the input files of
 * its transitive dependencies, and executes all affected scripts when they
 * change.
 *
 * Also watches all related package.json files and reloads script configuration
 * when they change.
 */
export class Watcher {
    constructor(rootScript, extraArgs, logger, workerPool, cache, failureMode, abort) {
        _Watcher_instances.add(this);
        /** See {@link WatcherState} */
        _Watcher_state.set(this, 'watching');
        _Watcher_rootScript.set(this, void 0);
        _Watcher_extraArgs.set(this, void 0);
        _Watcher_logger.set(this, void 0);
        _Watcher_workerPool.set(this, void 0);
        _Watcher_cache.set(this, void 0);
        _Watcher_failureMode.set(this, void 0);
        _Watcher_abort.set(this, void 0);
        /**
         * The most recent analysis of the root script. As soon as we detect it might
         * be stale because a package.json file was modified, this becomes undefined
         * again.
         */
        _Watcher_latestRootScriptConfig.set(this, void 0);
        /**
         * The file watcher for all package.json files relevant to this build graph.
         */
        _Watcher_configFilesWatcher.set(this, void 0);
        /**
         * File watchers for the input files of all scripts in this build graph.
         */
        _Watcher_inputFileWatchers.set(this, new Map());
        /**
         * Resolves when this watcher has been aborted and the last run finished.
         */
        _Watcher_finished.set(this, new Deferred());
        _Watcher_onConfigFileChanged.set(this, () => {
            __classPrivateFieldSet(this, _Watcher_latestRootScriptConfig, undefined, "f");
            __classPrivateFieldGet(this, _Watcher_fileChanged, "f").call(this);
        });
        _Watcher_fileChanged.set(this, () => {
            switch (__classPrivateFieldGet(this, _Watcher_state, "f")) {
                case 'watching': {
                    void __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_startRun).call(this);
                    return;
                }
                case 'running': {
                    __classPrivateFieldSet(this, _Watcher_state, 'queued', "f");
                    return;
                }
                case 'queued':
                case 'aborted': {
                    return;
                }
                default: {
                    throw unknownState(__classPrivateFieldGet(this, _Watcher_state, "f"));
                }
            }
        });
        __classPrivateFieldSet(this, _Watcher_rootScript, rootScript, "f");
        __classPrivateFieldSet(this, _Watcher_extraArgs, extraArgs, "f");
        __classPrivateFieldSet(this, _Watcher_logger, logger, "f");
        __classPrivateFieldSet(this, _Watcher_workerPool, workerPool, "f");
        __classPrivateFieldSet(this, _Watcher_failureMode, failureMode, "f");
        __classPrivateFieldSet(this, _Watcher_cache, cache, "f");
        __classPrivateFieldSet(this, _Watcher_abort, abort, "f");
    }
    static async watch(rootScript, extraArgs, logger, workerPool, cache, failureMode, abort) {
        const watcher = new Watcher(rootScript, extraArgs, logger, workerPool, cache, failureMode, abort);
        void __classPrivateFieldGet(watcher, _Watcher_instances, "m", _Watcher_startRun).call(watcher);
        void abort.promise.then(() => {
            __classPrivateFieldGet(watcher, _Watcher_instances, "m", _Watcher_onAbort).call(watcher);
        });
        return __classPrivateFieldGet(watcher, _Watcher_finished, "f").promise;
    }
}
_Watcher_state = new WeakMap(), _Watcher_rootScript = new WeakMap(), _Watcher_extraArgs = new WeakMap(), _Watcher_logger = new WeakMap(), _Watcher_workerPool = new WeakMap(), _Watcher_cache = new WeakMap(), _Watcher_failureMode = new WeakMap(), _Watcher_abort = new WeakMap(), _Watcher_latestRootScriptConfig = new WeakMap(), _Watcher_configFilesWatcher = new WeakMap(), _Watcher_inputFileWatchers = new WeakMap(), _Watcher_finished = new WeakMap(), _Watcher_onConfigFileChanged = new WeakMap(), _Watcher_fileChanged = new WeakMap(), _Watcher_instances = new WeakSet(), _Watcher_startRun = function _Watcher_startRun() {
    switch (__classPrivateFieldGet(this, _Watcher_state, "f")) {
        case 'watching':
        case 'queued': {
            __classPrivateFieldSet(this, _Watcher_state, 'running', "f");
            __classPrivateFieldGet(this, _Watcher_logger, "f").log({
                script: __classPrivateFieldGet(this, _Watcher_rootScript, "f"),
                type: 'info',
                detail: 'watch-run-start',
            });
            if (__classPrivateFieldGet(this, _Watcher_latestRootScriptConfig, "f") === undefined) {
                void __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_analyze).call(this);
            }
            else {
                void __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_execute).call(this, __classPrivateFieldGet(this, _Watcher_latestRootScriptConfig, "f"));
            }
            return;
        }
        case 'running':
        case 'aborted': {
            throw unexpectedState(__classPrivateFieldGet(this, _Watcher_state, "f"));
        }
        default: {
            throw unknownState(__classPrivateFieldGet(this, _Watcher_state, "f"));
        }
    }
}, _Watcher_analyze = async function _Watcher_analyze() {
    if (__classPrivateFieldGet(this, _Watcher_state, "f") !== 'running') {
        throw unexpectedState(__classPrivateFieldGet(this, _Watcher_state, "f"));
    }
    const analyzer = new Analyzer();
    const result = await analyzer.analyze(__classPrivateFieldGet(this, _Watcher_rootScript, "f"), __classPrivateFieldGet(this, _Watcher_extraArgs, "f"));
    if (__classPrivateFieldGet(this, _Watcher_state, "f") === 'aborted') {
        return;
    }
    // Set up watchers for all relevant config files even if there were errors
    // so that we'll try again when the user modifies a config file.
    const configFiles = [...result.relevantConfigFilePaths];
    // Order doesn't matter because we know we don't have any !negated patterns,
    // but we're going to compare arrays exactly so the order should be
    // deterministic.
    configFiles.sort();
    const oldWatcher = __classPrivateFieldGet(this, _Watcher_configFilesWatcher, "f");
    if (!watchPathsEqual(configFiles, oldWatcher?.patterns)) {
        __classPrivateFieldSet(this, _Watcher_configFilesWatcher, makeWatcher(configFiles, '/', __classPrivateFieldGet(this, _Watcher_onConfigFileChanged, "f")), "f");
        if (oldWatcher !== undefined) {
            void oldWatcher.watcher.close();
        }
    }
    if (!result.config.ok) {
        for (const error of result.config.error) {
            __classPrivateFieldGet(this, _Watcher_logger, "f").log(error);
        }
        __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_onRunDone).call(this);
        return;
    }
    __classPrivateFieldSet(this, _Watcher_latestRootScriptConfig, result.config.value, "f");
    __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_synchronizeInputFileWatchers).call(this, __classPrivateFieldGet(this, _Watcher_latestRootScriptConfig, "f"));
    void __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_execute).call(this, __classPrivateFieldGet(this, _Watcher_latestRootScriptConfig, "f"));
}, _Watcher_execute = async function _Watcher_execute(script) {
    if (__classPrivateFieldGet(this, _Watcher_state, "f") !== 'running') {
        throw unexpectedState(__classPrivateFieldGet(this, _Watcher_state, "f"));
    }
    const executor = new Executor(__classPrivateFieldGet(this, _Watcher_logger, "f"), __classPrivateFieldGet(this, _Watcher_workerPool, "f"), __classPrivateFieldGet(this, _Watcher_cache, "f"), __classPrivateFieldGet(this, _Watcher_failureMode, "f"), __classPrivateFieldGet(this, _Watcher_abort, "f"));
    const result = await executor.execute(script);
    if (!result.ok) {
        for (const error of result.error) {
            __classPrivateFieldGet(this, _Watcher_logger, "f").log(error);
        }
    }
    __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_onRunDone).call(this);
}, _Watcher_onRunDone = function _Watcher_onRunDone() {
    __classPrivateFieldGet(this, _Watcher_logger, "f").log({
        script: __classPrivateFieldGet(this, _Watcher_rootScript, "f"),
        type: 'info',
        detail: 'watch-run-end',
    });
    switch (__classPrivateFieldGet(this, _Watcher_state, "f")) {
        case 'queued': {
            void __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_startRun).call(this);
            return;
        }
        case 'running': {
            __classPrivateFieldSet(this, _Watcher_state, 'watching', "f");
            return;
        }
        case 'aborted': {
            __classPrivateFieldGet(this, _Watcher_finished, "f").resolve();
            return;
        }
        case 'watching': {
            throw unexpectedState(__classPrivateFieldGet(this, _Watcher_state, "f"));
        }
        default: {
            throw unknownState(__classPrivateFieldGet(this, _Watcher_state, "f"));
        }
    }
}, _Watcher_synchronizeInputFileWatchers = function _Watcher_synchronizeInputFileWatchers(root) {
    const visited = new Set();
    const visit = (script) => {
        const key = scriptReferenceToString(script);
        if (visited.has(key)) {
            return;
        }
        visited.add(key);
        const newInputFiles = script.files?.values;
        const oldWatcher = __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f").get(key);
        if (!watchPathsEqual(newInputFiles, oldWatcher?.patterns)) {
            if (newInputFiles === undefined || newInputFiles.length === 0) {
                __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f").delete(key);
            }
            else {
                const newWatcher = makeWatcher(newInputFiles, script.packageDir, __classPrivateFieldGet(this, _Watcher_fileChanged, "f"));
                __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f").set(key, newWatcher);
            }
            if (oldWatcher !== undefined) {
                void oldWatcher.watcher.close();
            }
        }
        for (const dep of script.dependencies) {
            visit(dep.config);
        }
    };
    visit(root);
    // There also could be some scripts that have been removed entirely.
    for (const [oldKey, oldWatcher] of __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f")) {
        if (!visited.has(oldKey)) {
            void oldWatcher.watcher.close();
            __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f").delete(oldKey);
        }
    }
}, _Watcher_onAbort = function _Watcher_onAbort() {
    switch (__classPrivateFieldGet(this, _Watcher_state, "f")) {
        case 'watching': {
            __classPrivateFieldSet(this, _Watcher_state, 'aborted', "f");
            __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_closeAllFileWatchers).call(this);
            __classPrivateFieldGet(this, _Watcher_finished, "f").resolve();
            return;
        }
        case 'running':
        case 'queued': {
            __classPrivateFieldSet(this, _Watcher_state, 'aborted', "f");
            __classPrivateFieldGet(this, _Watcher_instances, "m", _Watcher_closeAllFileWatchers).call(this);
            // Don't resolve #finished immediately so that we will wait for #analyze
            // or #execute to finish.
            return;
        }
        case 'aborted': {
            return;
        }
        default: {
            throw unknownState(__classPrivateFieldGet(this, _Watcher_state, "f"));
        }
    }
}, _Watcher_closeAllFileWatchers = function _Watcher_closeAllFileWatchers() {
    void __classPrivateFieldGet(this, _Watcher_configFilesWatcher, "f")?.watcher.close();
    for (const value of __classPrivateFieldGet(this, _Watcher_inputFileWatchers, "f").values()) {
        void value.watcher.close();
    }
};
const watchPathsEqual = (a, b) => {
    if (a === undefined && b === undefined) {
        return true;
    }
    if (a === undefined || b === undefined) {
        return false;
    }
    if (a.length !== b.length) {
        return false;
    }
    for (let i = 0; i < a.length; i++) {
        if (a[i] !== b[i]) {
            return false;
        }
    }
    return true;
};
const makeWatcher = (patterns, cwd, callback) => {
    const watcher = chokidar.watch(patterns, {
        cwd,
        // Ignore the initial "add" events emitted when chokidar first discovers
        // each file. We already do an initial run, so these events are just noise
        // that may trigger an unnecessary second run.
        // https://github.com/paulmillr/chokidar#path-filtering
        ignoreInitial: true,
    });
    watcher.on('all', callback);
    return {
        patterns,
        watcher,
    };
};
//# sourceMappingURL=watcher.js.map